import os
from typing import Union

import git
from git import InvalidGitRepositoryError, NoSuchPathError, Repo
from pydantic import BaseModel

from gto.base import BasePromotion, BaseRegistryState
from gto.config import CONFIG_FILE_NAME, RegistryConfig, read_registry_config
from gto.exceptions import (
    NoRepo,
    VersionAlreadyRegistered,
    VersionExistsForCommit,
    WrongArgs,
)
from gto.index import EnrichmentManager
from gto.tag import TagStageManager, TagVersionManager
from gto.ui import echo
from gto.versions import SemVer


class GitRegistry(BaseModel):
    repo: git.Repo
    version_manager: TagVersionManager
    stage_manager: TagStageManager
    enrichment_manager: EnrichmentManager
    config: RegistryConfig

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def from_repo(cls, repo=Union[str, Repo], config: RegistryConfig = None):
        if isinstance(repo, str):
            try:
                repo = git.Repo(repo, search_parent_directories=True)
            except (InvalidGitRepositoryError, NoSuchPathError) as e:
                raise NoRepo(repo) from e
        if config is None:
            config = read_registry_config(
                os.path.join(repo.working_dir, CONFIG_FILE_NAME)
            )

        return cls(
            repo=repo,
            config=config,
            version_manager=TagVersionManager(repo=repo, config=config),
            stage_manager=TagStageManager(repo=repo, config=config),
            enrichment_manager=EnrichmentManager(repo=repo, config=config),
        )

    def get_state(
        self,
        all_branches=False,
        all_commits=False,
    ) -> BaseRegistryState:
        state = BaseRegistryState()
        state = self.version_manager.update_state(state)
        state = self.stage_manager.update_state(state)
        state = self.enrichment_manager.update_state(
            state,
            all_branches=all_branches,
            all_commits=all_commits,
        )
        state.sort()
        return state

    def get_artifacts(
        self,
        all_branches=False,
        all_commits=False,
    ):
        return self.get_state(
            all_branches=all_branches,
            all_commits=all_commits,
        ).get_artifacts()

    def find_artifact(
        self,
        name: str = None,
        create_new=False,
        all_branches=False,
        all_commits=False,
    ):
        return self.get_state(
            all_branches=all_branches,
            all_commits=all_commits,
        ).find_artifact(
            name, create_new=create_new  # type: ignore
        )

    def register(
        self,
        name,
        ref,
        version=None,
        message=None,
        bump_major=False,
        bump_minor=False,
        bump_patch=False,
        stdout=False,
    ):
        """Register artifact version"""
        version_args = sum(
            bool(i) for i in (version, bump_major, bump_minor, bump_patch)
        )
        if version_args > 1:
            raise WrongArgs("Need to specify either version or single bump argument")
        ref = self.repo.commit(ref).hexsha
        # TODO: add the same check for other actions, to promote and etc
        # also we need to check integrity of the index+state
        found_artifact = self.find_artifact(name, create_new=True)
        # check that this commit don't have a version already
        found_version = found_artifact.find_version(commit_hexsha=ref)
        if found_version is not None and found_version.is_registered:
            raise VersionExistsForCommit(name, found_version.name)
        # if version name is provided, use it
        if version:
            if found_artifact.find_version(name=version) is not None:
                raise VersionAlreadyRegistered(version)
            if not SemVer.is_valid(version):
                raise WrongArgs(
                    f"Version '{version}' is not valid. Example of valid version: 'v1.0.0'"
                )
        else:
            # if version name wasn't provided but there were some, bump the last one
            last_version = found_artifact.get_latest_version(registered_only=True)
            if last_version:
                version = (
                    SemVer(last_version.name)
                    .bump(
                        bump_major=bump_major,
                        bump_minor=bump_minor,
                        bump_patch=bump_patch if (bump_major or bump_minor) else True,
                    )
                    .version
                )
            elif version_args:
                raise WrongArgs(
                    "Can't apply bump, because this is the first version being registered"
                )
            else:
                version = SemVer.get_minimal().version
        self.version_manager.register(
            name,
            version,
            ref,
            message=message or f"Registering artifact {name} version {version}",
        )
        registered_version = self.find_artifact(name).find_version(
            name=version, raise_if_not_found=True
        )
        if stdout:
            echo(
                f"Created git tag '{registered_version.tag}' that registers a new version"
            )
        return registered_version

    def promote(
        self,
        name,
        stage,
        promote_version=None,
        promote_ref=None,
        name_version=None,
        message=None,
        simple=False,
        force=False,
        skip_registration=False,
        stdout=False,
    ) -> BasePromotion:
        """Assign stage to specific artifact version"""
        self.config.assert_stage(stage)
        if not (promote_version is None) ^ (promote_ref is None):
            raise WrongArgs("One and only one of (version, ref) must be specified.")
        if name_version and skip_registration:
            raise WrongArgs(
                "You either need to supply version name or skip registration"
            )
        if promote_ref:
            promote_ref = self.repo.commit(promote_ref).hexsha
        found_artifact = self.find_artifact(name, create_new=True)
        if promote_version:
            found_version = found_artifact.find_version(
                name=promote_version, raise_if_not_found=False
            )
            if not found_version:
                raise WrongArgs(f"Version '{promote_version}' is not registered")
            promote_ref = self.find_commit(name, promote_version)
        else:
            found_version = found_artifact.find_version(commit_hexsha=promote_ref)
            if found_version:
                if name_version:
                    raise WrongArgs(
                        f"Can't register '{SemVer(name_version).version}', since '{found_version.name}' is registered already at this ref"
                    )
            elif not skip_registration:
                self.register(
                    name, version=name_version, ref=promote_ref, stdout=stdout
                )
        if (
            not force
            and found_version
            and found_version.stage
            and found_version.stage.stage == stage
        ):
            raise WrongArgs(f"Version is already in stage '{stage}'")
        self.stage_manager.promote(  # type: ignore
            name,
            stage,
            ref=promote_ref,
            message=message
            or f"Promoting {name} version {promote_version} to stage {stage}",
            simple=simple,
        )
        promotion = (
            self.get_state()
            .find_artifact(name)
            .find_version_at_commit(promote_ref)
            .stage
        )
        if stdout:
            echo(
                f"Created git tag '{promotion.tag}' that promotes '{promotion.version}'"
            )
        return promotion

    def check_ref(self, ref: str):
        "Find out what was registered/promoted in this ref"
        return {
            "version": self.version_manager.check_ref(ref, self.get_state()),
            "stage": self.stage_manager.check_ref(ref, self.get_state()),
        }

    def find_commit(self, name, version):
        return self.get_state().find_commit(name, version)

    def which(
        self, name, stage, raise_if_not_found=True, all=False, registered_only=False
    ):
        """Return stage active in specific stage"""
        return self.get_state().which(
            name, stage, raise_if_not_found, all=all, registered_only=registered_only
        )

    def latest(self, name: str, all: bool = False, registered: bool = True):
        """Return latest active version for artifact"""
        artifact = self.get_state().find_artifact(name)
        if all:
            return artifact.sort_versions(registered=registered)
        return artifact.get_latest_version(registered_only=registered)

    def get_stages(self, allowed: bool = False):
        """Return list of stages in the registry.
        If "allowed", return stages that are allowed in config.
        """
        if allowed:
            return self.config.stages
        return sorted(
            {stage for o in self.get_artifacts().values() for stage in o.unique_stages}
        )



from typing import Any, Dict, Sequence, Set, Union

from pydantic import BaseModel

# def _equals(a, b):
#     # separate function is helpful for debug
#     # cause you see dicts without skip_keys
#     assert a == b


def _check_obj(
    obj: BaseModel, values: Dict[str, Any], skip_keys: Union[Set[str], Sequence[str]]
):
    obj_values = obj.dict(exclude=set(skip_keys))
    assert obj_values == values


def _check_dict(
    obj: Dict[str, Any],
    values: Dict[str, Any],
    skip_keys: Union[Set[str], Sequence[str]],
):
    obj_values = {k: v for k, v in obj.items() if k not in skip_keys}
    values = {k: v for k, v in values.items() if k not in skip_keys}
    assert obj_values == values



from collections import OrderedDict
from datetime import datetime
from typing import List, Optional, Union

from git import Repo

from gto.constants import NAME, STAGE, VERSION
from gto.exceptions import WrongArgs
from gto.ext import EnrichmentInfo
from gto.index import (
    EnrichmentManager,
    FileIndexManager,
    RepoIndexManager,
    init_index_manager,
)
from gto.registry import GitRegistry
from gto.tag import NAME_REFERENCE
from gto.tag import parse_name as parse_tag_name
from gto.tag import parse_name_reference


def get_index(repo: Union[str, Repo], file=False):
    """Get index state"""
    if file:
        return FileIndexManager.from_path(
            path=repo if isinstance(repo, str) else repo.working_dir
        )
    return RepoIndexManager.from_repo(repo)


def _get_state(repo: Union[str, Repo]):
    """Show current registry state"""
    return GitRegistry.from_repo(repo).get_state()


def get_stages(repo: Union[str, Repo], allowed: bool = False):
    return GitRegistry.from_repo(repo).get_stages(allowed=allowed)


# TODO: make this work the same as CLI version
def annotate(
    repo: Union[str, Repo],
    name: str,
    type: Optional[str] = None,
    path: Optional[str] = None,
    must_exist: bool = False,
    labels: List[str] = None,
    description: str = "",
    # update: bool = False,
):
    """Add an artifact to the Index"""
    return init_index_manager(path=repo).add(
        name,
        type=type,
        path=path,
        must_exist=must_exist,
        labels=labels,
        description=description,
        update=True,
    )


def remove(repo: Union[str, Repo], name: str):
    """Remove an artifact from the Index"""
    return init_index_manager(path=repo).remove(name)


def register(
    repo: Union[str, Repo],
    name: str,
    ref: str,
    version: str = None,
    message: str = None,
    bump_major: bool = False,
    bump_minor: bool = False,
    bump_patch: bool = False,
    stdout: bool = False,
):
    """Register new artifact version"""
    return GitRegistry.from_repo(repo).register(
        name=name,
        ref=ref,
        version=version,
        message=message,
        bump_major=bump_major,
        bump_minor=bump_minor,
        bump_patch=bump_patch,
        stdout=stdout,
    )


def promote(
    repo: Union[str, Repo],
    name: str,
    stage: str,
    promote_version: str = None,
    promote_ref: str = None,
    name_version: str = None,
    message: str = None,
    simple: bool = False,
    force: bool = False,
    skip_registration: bool = False,
    stdout: bool = False,
):
    """Assign stage to specific artifact version"""
    return GitRegistry.from_repo(repo).promote(
        name,
        stage,
        promote_version,
        promote_ref,
        name_version,
        message=message,
        simple=simple,
        force=force,
        skip_registration=skip_registration,
        stdout=stdout,
    )


def parse_tag(name: str):
    return parse_tag_name(name)


def find_latest_version(
    repo: Union[str, Repo],
    name: str,
    all: bool = False,
    registered: bool = True,
):
    """Return latest version for artifact"""
    return GitRegistry.from_repo(repo).latest(name, all=all, registered=registered)


def find_versions_in_stage(
    repo: Union[str, Repo],
    name: str,
    stage: str,
    all: bool = False,
    registered_only: bool = False,
):
    """Return version of artifact with specific stage active"""
    return GitRegistry.from_repo(repo).which(
        name, stage, raise_if_not_found=False, all=all, registered_only=registered_only
    )


def check_ref(repo: Union[str, Repo], ref: str):
    """Find out what have been registered/promoted in the provided ref"""
    reg = GitRegistry.from_repo(repo)
    if ref.startswith("refs/tags/"):
        ref = ref[len("refs/tags/") :]
    if ref.startswith("refs/heads/"):
        ref = reg.repo.commit(ref).hexsha
    result = reg.check_ref(ref)
    return {
        action: {name: version.dict() for name, version in found.items()}
        for action, found in result.items()
    }


def show(
    repo: Union[str, Repo],
    name: Optional[str] = None,
    table: bool = False,
    all_branches=False,
    all_commits=False,
    truncate_hexsha=False,
    registered_only=False,
):
    return (
        _show_versions(
            repo,
            name=name,
            all_branches=all_branches,
            all_commits=all_commits,
            registered_only=registered_only,
            table=table,
            truncate_hexsha=truncate_hexsha,
        )
        if name
        else _show_registry(
            repo,
            all_branches=all_branches,
            all_commits=all_commits,
            registered_only=registered_only,
            table=table,
            truncate_hexsha=truncate_hexsha,
        )
    )


def _show_registry(
    repo: Union[str, Repo],
    all_branches=False,
    all_commits=False,
    registered_only=False,
    table: bool = False,
    truncate_hexsha: bool = False,  # pylint: disable=unused-argument
):
    """Show current registry state"""

    def format_hexsha(hexsha):
        return hexsha[:7] if truncate_hexsha else hexsha

    reg = GitRegistry.from_repo(repo)
    stages = list(reg.get_stages())
    models_state = {
        o.name: {
            "version": format_hexsha(o.get_latest_version(registered_only=True).name)
            if o.get_latest_version(registered_only=True)
            else None,
            "stage": {
                name: format_hexsha(
                    o.get_promotions(registered_only=registered_only)[name].version
                )
                if name in o.get_promotions(registered_only=registered_only)
                else None
                for name in stages
            },
        }
        for o in reg.get_artifacts(
            all_branches=all_branches,
            all_commits=all_commits,
        ).values()
    }
    if not table:
        return models_state

    result = [
        [name, d["version"]] + [d["stage"][name] for name in stages]
        for name, d in models_state.items()
    ]
    headers = ["name", "latest"] + [f"#{e}" for e in stages]
    return result, headers


def _show_versions(
    repo: Union[str, Repo],
    name: str,
    raw: bool = False,
    all_branches=False,
    all_commits=False,
    registered_only=False,
    table: bool = False,
    truncate_hexsha: bool = False,
):
    """List versions of artifact"""

    def format_hexsha(hexsha):
        return hexsha[:7] if truncate_hexsha else hexsha

    reg = GitRegistry.from_repo(repo)
    if raw:
        return reg.find_artifact(name).versions
    versions = [
        v.dict_status()
        for v in reg.find_artifact(
            name,
            all_branches=all_branches,
            all_commits=all_commits,
        ).get_versions(
            include_non_explicit=not registered_only, include_discovered=True
        )
    ]
    if not table:
        return versions

    first_keys = ["artifact", "version", "stage"]
    versions_ = []
    for v in versions:
        v["version"] = format_hexsha(v["name"])
        if v["stage"]:
            v["stage"] = v["stage"]["stage"]
        v["commit_hexsha"] = format_hexsha(v["commit_hexsha"])
        v["ref"] = v["tag"] or v["commit_hexsha"]
        for key in (
            "enrichments",
            "discovered",
            "tag",
            "commit_hexsha",
            "name",
            "message",
            "author_email",
        ):
            v.pop(key)
        # v["enrichments"] = [e["source"] for e in v["enrichments"]]
        v = OrderedDict(
            [(key, v[key]) for key in first_keys]
            + [(key, v[key]) for key in v if key not in first_keys]
        )
        versions_.append(v)
    return versions_, "keys"


def describe(
    repo: Union[str, Repo], name: str, rev: str = None
) -> List[EnrichmentInfo]:
    """Find enrichments for the artifact"""
    ref_type, parsed = parse_name_reference(name)
    if ref_type == NAME_REFERENCE.NAME:
        return EnrichmentManager.from_repo(repo).describe(name=name, rev=rev)
    if ref_type == NAME_REFERENCE.TAG:
        if rev:
            raise WrongArgs("Should not specify revision if you pass git tag")
        return EnrichmentManager.from_repo(repo).describe(name=parsed[NAME], rev=name)
    raise NotImplementedError


def history(  # pylint: disable=too-many-locals
    repo: Union[str, Repo],
    artifact: str = None,
    # action: str = None,
    all_branches=False,
    all_commits=False,
    ascending: bool = False,
    table: bool = False,
    truncate_hexsha: bool = False,
):

    reg = GitRegistry.from_repo(repo)
    artifacts = reg.get_artifacts(
        all_branches=all_branches,
        all_commits=all_commits,
    )

    def format_hexsha(hexsha):
        return hexsha[:7] if truncate_hexsha else hexsha

    commits = [
        OrderedDict(
            timestamp=datetime.fromtimestamp(
                reg.repo.commit(v.commit_hexsha).committed_date
            ),
            artifact=o.name,
            event="commit",
            commit=format_hexsha(v.commit_hexsha),
            author=reg.repo.commit(v.commit_hexsha).author.name,
            author_email=reg.repo.commit(v.commit_hexsha).author.email,
            message=reg.repo.commit(v.commit_hexsha).message,
        )
        for o in artifacts.values()
        for v in o.get_versions(include_non_explicit=True, include_discovered=True)
    ]

    registration = [
        OrderedDict(
            timestamp=v.created_at,
            artifact=o.name,
            event="registration",
            version=format_hexsha(v.name),
            commit=format_hexsha(v.commit_hexsha),
            author=v.author,
            author_email=v.author_email,
            message=v.message,
            # enrichments=[e.source for e in v.enrichments],
        )
        for o in artifacts.values()
        for v in o.get_versions()
    ]

    promotion = [
        OrderedDict(
            timestamp=l.created_at,
            artifact=o.name,
            event="promotion",
            version=format_hexsha(l.version),
            stage=l.stage,
            commit=format_hexsha(l.commit_hexsha),
            author=l.author,
            author_email=l.author_email,
            message=l.message,
        )
        for o in artifacts.values()
        for l in o.stages
    ]

    events_order = {
        "commit": 1,
        "registration": 2,
        "promotion": 3,
    }
    events = sorted(
        commits + registration + promotion,
        key=lambda x: (x["timestamp"], events_order[x["event"]]),
    )
    if not ascending:
        events.reverse()
    if artifact:
        events = [event for event in events if event["artifact"] == artifact]
    if not table:
        return events
    keys_order = [
        "timestamp",
        "artifact",
        "event",
        VERSION,
        STAGE,
        # "enrichments",
        "commit",
        "author",
    ]
    keys_order = [c for c in keys_order if any(c in event for event in events)]
    events = [
        OrderedDict((key, event.get(key)) for key in keys_order) for event in events
    ]
    return events, "keys"



from typing import Callable, Tuple

import git
import pytest

from gto import CONFIG
from gto.index import (
    Artifact,
    RepoIndexManager,
    check_if_path_exists,
    find_repeated_path,
)


def init_index(path):
    return RepoIndexManager.from_repo(path)


@pytest.fixture
def git_index_repo(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo  # pylint: disable=unused-variable
    return init_index(repo.working_dir), repo


def test_git_index_add_virtual(git_index_repo):
    index, repo = git_index_repo
    index.add(
        "a",
        type="a",
        path="a",
        must_exist=False,
        labels=[],
        description="",
        update=False,
    )

    new_index = init_index(repo.git_dir)
    assert isinstance(new_index, RepoIndexManager)
    index_value = new_index.get_index()
    assert index_value.state["a"] == Artifact(path="a", type="a", virtual=True)

    repo.index.add(CONFIG.INDEX)
    commit = repo.index.commit("add index")

    assert new_index.get_history()[commit.hexsha].state == index_value.state


def test_git_index_remove_virtual(git_index_repo):
    index, repo = git_index_repo
    index.add("a", "a", "a", must_exist=False, labels=[], description="", update=True)

    new_index = init_index(repo.git_dir)
    assert isinstance(new_index, RepoIndexManager)

    new_index.remove("a")
    index_value = new_index.get_index()
    assert index_value.state == {}


@pytest.mark.parametrize(
    "path, paths",
    [
        ("models/m1.txt", ["models/m1.txt"]),
        ("models/m1.txt", ["models"]),
        ("models", ["models/m1.txt"]),
        ("models", ["models/"]),
        ("models/m1.txt/", ["models/m1.txt"]),
    ],
)
def test_check_path_found(path, paths):
    assert find_repeated_path(path, paths) is not None


@pytest.mark.parametrize(
    "path, paths",
    [("models/m1.txt", ["models/m2.txt"]), ("models/a/m1.txt", ["models/m1.txt"])],
)
def test_check_path_not_found(path, paths):
    assert find_repeated_path(path, paths) is None


def test_check_existence_repo(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo

    write_file("m1.txt", "some content")
    repo.index.add(["m1.txt"])
    repo.index.commit("commit config file")
    write_file("a/b", "some content")
    repo.index.add(["a/b"])
    repo.index.remove(["m1.txt"])
    repo.index.commit("commit a/b")

    assert check_if_path_exists("m1.txt", repo, "HEAD^1")
    assert not check_if_path_exists("m1.txt", repo, "HEAD")
    assert not check_if_path_exists("a/b", repo, "HEAD^1")
    assert check_if_path_exists("a/b", repo, "HEAD")


def test_check_existence_no_repo(tmp_path):
    with open(tmp_path / "m1.txt", "w", encoding="utf8") as f:
        f.write("some content")
    assert check_if_path_exists(tmp_path / "m1.txt")
    assert not check_if_path_exists(tmp_path / "not/exists")



import os
from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime
from functools import wraps
from pathlib import Path
from typing import IO, Dict, FrozenSet, Generator, List, Optional, Union

import git
from pydantic import BaseModel, parse_obj_as

from gto.base import BaseManager, BaseRegistryState, BaseVersion
from gto.config import (
    CONFIG_FILE_NAME,
    RegistryConfig,
    read_registry_config,
    yaml,
)
from gto.constants import Action
from gto.exceptions import (
    ArtifactExists,
    ArtifactNotFound,
    NoFile,
    NoRepo,
    PathIsUsed,
    WrongArgs,
)
from gto.ext import Enrichment, EnrichmentInfo
from gto.utils import resolve_ref


class Artifact(BaseModel):
    type: Optional[str] = None
    path: Optional[str] = None
    virtual: bool = True
    labels: List[str] = []  # TODO: allow key:value labels
    description: str = ""


State = Dict[str, Artifact]


def not_frozen(func):
    @wraps(func)
    def inner(self: "Index", *args, **kwargs):
        if self.frozen:
            raise ValueError(f"Cannot {func.__name__}: {self.__class__} is frozen")
        return func(self, *args, **kwargs)

    return inner


def find_repeated_path(
    path: Union[str, Path], paths: List[Union[str, Path]]
) -> Optional[Path]:
    """Return the path from "paths" that conflicts with "path":
    is equal to it or is a subpath (in both directions).
    """
    path = Path(path).resolve()
    for p in paths:
        p = Path(p).resolve()
        if p == path or p in path.parents or path in p.parents:
            return p
    return None


def check_if_path_exists(path: str, repo: git.Repo = None, ref: str = None):
    """Check if path was committed to repo
    or it just exists in case the repo is not provided.
    """
    if repo is None:
        return Path(path).exists()
    try:
        _ = (repo.commit(ref).tree / path).data_stream
        return True
    except KeyError:
        return False


def traverse_commit(commit: git.Commit) -> Generator[git.Commit, None, None]:
    yield commit
    for parent in commit.parents:
        yield from traverse_commit(parent)


class Index(BaseModel):
    state: State = {}  # TODO should not be populated until load() is called
    frozen: bool = False

    def __contains__(self, item):
        return item in self.state

    @classmethod
    def read(cls, path_or_file: Union[str, IO], frozen: bool = False):
        index = cls(frozen=frozen)
        index.state = index.read_state(path_or_file)
        return index

    @staticmethod
    def read_state(path_or_file: Union[str, IO]):
        if isinstance(path_or_file, str):
            with open(path_or_file, "r", encoding="utf8") as file:
                return parse_obj_as(State, yaml.load(file))
        return parse_obj_as(State, yaml.load(path_or_file))

    def write_state(self, path_or_file: Union[str, IO]):
        if isinstance(path_or_file, str):
            with open(path_or_file, "w", encoding="utf8") as file:
                state = self.dict(exclude_defaults=True).get("state", {})
                yaml.dump(state, file)

    @not_frozen
    def add(
        self, name, type, path, must_exist, labels, description, update
    ) -> Artifact:
        if name in self and not update:
            raise ArtifactExists(name)
        if (
            must_exist
            and find_repeated_path(
                path, [a.path for n, a in self.state.items() if n != name]
            )
            is not None
        ):
            raise PathIsUsed(type=type, name=name, path=path)
        if update and name in self.state:
            self.state[name].type = type or self.state[name].type
            self.state[name].path = path or self.state[name].path
            if must_exist:
                self.state[name].virtual = False
            elif path:
                self.state[name].virtual = True
            self.state[name].labels = sorted(set(self.state[name].labels).union(labels))
            self.state[name].description = description or self.state[name].description
        else:
            self.state[name] = Artifact(
                type=type,
                path=path,
                virtual=not must_exist,
                labels=labels,
                description=description,
            )
        return self.state[name]

    @not_frozen
    def remove(self, name):
        if name not in self:
            raise ArtifactNotFound(name)
        del self.state[name]


class BaseIndexManager(BaseModel, ABC):
    current: Optional[Index]
    config: RegistryConfig

    @abstractmethod
    def get_index(self) -> Index:
        raise NotImplementedError

    @abstractmethod
    def update(self):
        raise NotImplementedError

    @abstractmethod
    def get_history(self) -> Dict[str, Index]:
        raise NotImplementedError

    def add(self, name, type, path, must_exist, labels, description, update):
        if type:
            self.config.assert_type(type)
        if must_exist:
            if not path:
                raise WrongArgs("`path` is required when `must_exist` is set to True")
            if not check_if_path_exists(
                path, self.repo if hasattr(self, "repo") else None
            ):
                raise NoFile(path)
        index = self.get_index()
        index.add(
            name,
            type=type,
            path=path,
            must_exist=must_exist,
            labels=labels or [],
            description=description,
            update=update,
        )
        self.update()

    def remove(self, name):
        index = self.get_index()
        index.remove(name)
        self.update()


class FileIndexManager(BaseIndexManager):
    path: str = ""

    @classmethod
    def from_path(cls, path: str, config: RegistryConfig = None):
        if config is None:
            config = read_registry_config(os.path.join(path, CONFIG_FILE_NAME))
        return cls(path=path, config=config)

    def index_path(self):
        return str(Path(self.path) / self.config.INDEX)

    def get_index(self) -> Index:
        if os.path.exists(self.index_path()):
            self.current = Index.read(self.index_path())
        if not self.current:
            self.current = Index()
        return self.current

    def update(self):
        if self.current is not None:
            self.current.write_state(self.index_path())

    def get_history(self) -> Dict[str, Index]:
        raise NotImplementedError("Not a git repo: history is not available")


ArtifactCommits = Dict[str, Artifact]
ArtifactsCommits = Dict[str, ArtifactCommits]


class RepoIndexManager(FileIndexManager):
    repo: git.Repo

    @classmethod
    def from_repo(cls, repo: Union[str, git.Repo], config: RegistryConfig = None):
        if isinstance(repo, str):
            try:
                repo = git.Repo(repo, search_parent_directories=True)
            except git.InvalidGitRepositoryError as e:
                raise NoRepo(repo) from e
        if config is None:
            config = read_registry_config(
                os.path.join(repo.working_dir, CONFIG_FILE_NAME)
            )
        return cls(repo=repo, config=config)

    def index_path(self):
        # TODO: config should be loaded from repo too
        return os.path.join(os.path.dirname(self.repo.git_dir), self.config.INDEX)

    class Config:
        arbitrary_types_allowed = True

    def get_commit_index(
        self, ref: Union[str, git.Reference, None], allow_to_not_exist: bool = True
    ) -> Optional[Index]:
        if not ref or isinstance(ref, str):
            ref = resolve_ref(self.repo, ref)
        if self.config.INDEX in ref.tree:
            return Index.read(
                (ref.tree / self.config.INDEX).data_stream,
                frozen=True,
            )
        if allow_to_not_exist:
            return None
        raise ValueError(f"No Index exists at {ref}")

    def get_history(self) -> Dict[str, Index]:
        commits = {
            commit
            for branch in self.repo.heads
            for commit in traverse_commit(branch.commit)
        }
        return {
            commit.hexsha: self.get_commit_index(commit.hexsha)  # type: ignore
            for commit in commits
            if self.config.INDEX in commit.tree
        }

    def artifact_centric_representation(self) -> ArtifactsCommits:
        representation = defaultdict(dict)  # type: ArtifactsCommits
        for commit, index in self.get_history().items():
            for art in index.state:
                representation[art][commit] = index.state[art]
        return representation

    def check_existence(self, name, commit):
        return name in self.get_commit_index(commit)

    def assert_existence(self, name, commit):
        if not self.check_existence(name, commit):
            raise ArtifactNotFound(name)


class EnrichmentManager(BaseManager):
    actions: FrozenSet[Action] = frozenset()

    @classmethod
    def from_repo(cls, repo, config: RegistryConfig = None):
        if isinstance(repo, str):
            repo = git.Repo(repo, search_parent_directories=True)
        if config is None:
            config = read_registry_config(
                os.path.join(repo.working_dir, CONFIG_FILE_NAME)
            )
        return cls(repo=repo, config=config)

    def describe(self, name: str, rev: str = None) -> List[EnrichmentInfo]:
        enrichments = self.config.enrichments
        res = []
        gto_enrichment = enrichments.pop("gto")
        gto_info = gto_enrichment.describe(self.repo, name, rev)
        if gto_info:
            res.append(gto_info)
            path = gto_info.get_path()  # type: ignore
            for enrichment in enrichments.values():
                enrichment_data = enrichment.describe(self.repo, path, rev)
                if enrichment_data is not None:
                    res.append(enrichment_data)
        return res

    def get_commits(self, all_branches=False, all_commits=False):
        if not self.repo.refs:
            return {}
        if all_commits:
            return {
                commit
                for branch in self.repo.heads
                for commit in traverse_commit(branch.commit)
            }
        if all_branches:
            return {branch.commit for branch in self.repo.heads}
        return {self.repo.commit()}

    def update_state(
        self,
        state: BaseRegistryState,
        all_branches=False,
        all_commits=False,
    ) -> BaseRegistryState:
        # processing registered artifacts and versions first
        for artifact in state.get_artifacts().values():
            for version in artifact.versions:
                enrichments = self.describe(artifact.name, rev=version.commit_hexsha)
                artifact.update_enrichments(
                    version=version.name, enrichments=enrichments
                )
                state.update_artifact(artifact)
        for commit in self.get_commits(
            all_branches=all_branches, all_commits=all_commits
        ):
            for art_name in GTOEnrichment().discover(self.repo, commit):
                enrichments = self.describe(art_name, rev=commit)
                artifact = state.find_artifact(art_name, create_new=True)
                version = artifact.find_version(commit_hexsha=commit.hexsha)
                # TODO: duplicated in tag.py
                if version:
                    version = version.name
                else:
                    artifact.add_version(
                        BaseVersion(
                            artifact=art_name,
                            name=commit.hexsha,
                            created_at=datetime.fromtimestamp(commit.committed_date),
                            author=commit.author.name,
                            author_email=commit.author.email,
                            commit_hexsha=commit.hexsha,
                            discovered=True,
                        )
                    )
                    version = commit.hexsha
                artifact.update_enrichments(version=version, enrichments=enrichments)
                state.update_artifact(artifact)
        return state

    def check_ref(self, ref: str, state: BaseRegistryState):
        # TODO: implement
        raise NotImplementedError()


def init_index_manager(path):
    try:
        return RepoIndexManager.from_repo(path)
    except NoRepo:
        return FileIndexManager.from_path(path)


class GTOInfo(EnrichmentInfo):
    source = "gto"
    artifact: Artifact

    def get_object(self) -> BaseModel:
        return self.artifact

    def get_human_readable(self) -> str:
        return self.artifact.json()

    def get_path(self):
        return self.artifact.path


class GTOEnrichment(Enrichment):
    source = "gto"

    def discover(  # pylint: disable=no-self-use
        self, repo, rev: Optional[str]
    ) -> Dict[str, GTOInfo]:
        index = RepoIndexManager.from_repo(repo).get_commit_index(rev)
        if index:
            return {
                name: GTOInfo(artifact=artifact)
                for name, artifact in index.state.items()
            }
        return {}

    def describe(  # pylint: disable=no-self-use
        self, repo, obj: str, rev: Optional[str]
    ) -> Optional[GTOInfo]:
        index = RepoIndexManager.from_repo(repo).get_commit_index(rev)
        if index and obj in index.state:
            return GTOInfo(artifact=index.state[obj])
        return None



from enum import Enum

REF = "ref"
# COMMIT = "commit"
TAG = "tag"
# BRANCH = "branch"
# FILE = "file"

ACTION = "action"
TYPE = "type"
NAME = "name"
PATH = "path"
VERSION = "version"
STAGE = "stage"
NUMBER = "number"


class Action(Enum):
    REGISTER = "register"
    PROMOTE = "promote"


class VersionSort(Enum):
    SemVer = "semver"
    Timestamp = "timestamp"



from abc import ABC, abstractmethod
from functools import lru_cache
from importlib import import_module
from typing import Dict, Optional, Type

import entrypoints
from pydantic import BaseModel

ENRICHMENT_ENRTYPOINT = "gto.enrichment"


def import_string(path):
    split = path.split(".")
    module_name, object_name = ".".join(split[:-1]), split[-1]
    mod = import_module(module_name)
    try:
        return getattr(mod, object_name)
    except AttributeError as e:
        raise ImportError(f"No object {object_name} in module {module_name}") from e


class EnrichmentInfo(BaseModel, ABC):
    source: str

    @abstractmethod
    def get_object(self) -> BaseModel:
        raise NotImplementedError

    def get_dict(self):
        return self.get_object().dict()

    @abstractmethod
    def get_human_readable(self) -> str:
        raise NotImplementedError


class Enrichment(BaseModel, ABC):
    source: str

    @abstractmethod
    def describe(
        self, repo: str, obj: str, rev: Optional[str]
    ) -> Optional[EnrichmentInfo]:
        raise NotImplementedError


# class CLIEnrichmentInfo(EnrichmentInfo):
#     data: Dict
#     repr: str

#     def get_object(self) -> BaseModel:
#         return self

#     def get_human_readable(self) -> str:
#         return self.repr


# class CLIEnrichment(Enrichment):
#     cmd: str
#     info_type: Union[str, Type[EnrichmentInfo]] = CLIEnrichmentInfo

#     @validator("info_type")
#     def info_class_validator(
#         cls, value
#     ):  # pylint: disable=no-self-argument,no-self-use  # noqa: B902
#         if isinstance(value, type):
#             return value
#         info_class = import_string(value)
#         if not isinstance(info_class, type) or not issubclass(
#             info_class, EnrichmentInfo
#         ):
#             raise ValueError(
#                 "Wrong value for info_type: should be class or string path to class (e.g. `package.module.ClassName`)"
#             )
#         return info_class

#     @property
#     def info_class(self) -> Type[EnrichmentInfo]:
#         return self.info_class_validator(self.info_type)

#     def describe(self, obj: str) -> Optional[EnrichmentInfo]:
#         try:
#             data = loads(subprocess.check_output(self.cmd.split() + [obj]))
#             return parse_obj_as(self.info_class, data)
#         except subprocess.SubprocessError:
#             return None


@lru_cache()
def _find_enrichments():
    eps = entrypoints.get_group_named(ENRICHMENT_ENRTYPOINT)
    return {k: ep.load() for k, ep in eps.items()}


@lru_cache()
def find_enrichments() -> Dict[str, Enrichment]:
    enrichments = _find_enrichments()
    res = {}
    for name, e in enrichments.items():
        # if isinstance(e, type) and issubclass(e, Enrichment) and not e.__fields_set__:
        if isinstance(e, type) and issubclass(e, Enrichment):
            res[name] = e()
        if isinstance(e, Enrichment):
            res[name] = e
    return res


@lru_cache()
def find_enrichment_types() -> Dict[str, Type[Enrichment]]:
    enrichments = _find_enrichments()
    return {
        k: e
        for k, e in enrichments.items()
        if isinstance(e, type) and issubclass(e, Enrichment)
    }



# pylint: disable=no-self-use, no-self-argument, inconsistent-return-statements, invalid-name, import-outside-toplevel
import re
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, BaseSettings, validator
from pydantic.env_settings import InitSettingsSource
from ruamel.yaml import YAML

from gto.exceptions import (
    UnknownStage,
    UnknownType,
    ValidationError,
    WrongConfig,
)
from gto.ext import Enrichment, find_enrichment_types, find_enrichments

yaml = YAML(typ="safe", pure=True)
yaml.default_flow_style = False

CONFIG_FILE_NAME = ".gto"


def check_name_is_valid(name):
    return bool(re.match(r"[a-zA-Z0-9-/]*$", name))


def assert_name_is_valid(name):
    if not check_name_is_valid(name):
        raise ValidationError(
            f"Invalid value '{name}'. Only alphanumeric characters, '-', '/' are allowed."
        )


class EnrichmentConfig(BaseModel):
    type: str
    config: Dict = {}

    def load(self) -> Enrichment:
        return find_enrichment_types()[self.type](**self.config)


class NoFileConfig(BaseSettings):
    INDEX: str = "artifacts.yaml"
    TYPE_ALLOWED: List[str] = []
    STAGE_ALLOWED: List[str] = []
    LOG_LEVEL: str = "INFO"
    DEBUG: bool = False
    ENRICHMENTS: List[EnrichmentConfig] = []
    AUTOLOAD_ENRICHMENTS: bool = True
    CONFIG_FILE_NAME: Optional[str] = CONFIG_FILE_NAME
    EMOJIS: bool = True

    class Config:
        env_prefix = "gto_"

    def assert_type(self, name):
        assert_name_is_valid(name)
        if self.TYPE_ALLOWED and name not in self.TYPE_ALLOWED:
            raise UnknownType(name, self.TYPE_ALLOWED)

    def assert_stage(self, name):
        assert_name_is_valid(name)
        if self.stages and name not in self.stages:
            raise UnknownStage(name, self.stages)

    @property
    def enrichments(self) -> Dict[str, Enrichment]:
        res = {e.source: e for e in (e.load() for e in self.ENRICHMENTS)}
        if self.AUTOLOAD_ENRICHMENTS:
            return {**find_enrichments(), **res}
        return res

    @property
    def stages(self) -> List[str]:
        return self.STAGE_ALLOWED

    @validator("TYPE_ALLOWED")
    def types_are_valid(cls, v):
        for name in v:
            assert_name_is_valid(name)
        return v

    @validator("STAGE_ALLOWED")
    def stages_are_valid(cls, v):
        for name in v:
            assert_name_is_valid(name)
        return v


def _set_location_init_source(init_source: InitSettingsSource):
    def inner(settings: "RegistryConfig"):
        if "CONFIG_FILE_NAME" in init_source.init_kwargs:
            settings.__dict__["CONFIG_FILE_NAME"] = init_source.init_kwargs[
                "CONFIG_FILE_NAME"
            ]
        return {}

    return inner


def config_settings_source(settings: "RegistryConfig") -> Dict[str, Any]:
    """
    A simple settings source that loads variables from a yaml file in GTO DIR
    """

    encoding = settings.__config__.env_file_encoding
    config_file = getattr(settings, "CONFIG_FILE_NAME", CONFIG_FILE_NAME)
    if not isinstance(config_file, Path):
        config_file = Path(config_file)
    if not config_file.exists():
        return {}
    conf = yaml.load(config_file.read_text(encoding=encoding))

    return {k.upper(): v for k, v in conf.items()} if conf else {}


class RegistryConfig(NoFileConfig):
    class Config:
        env_prefix = "gto_"
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                _set_location_init_source(init_settings),
                init_settings,
                env_settings,
                config_settings_source,
                file_secret_settings,
            )


def read_registry_config(config_file_name):
    try:
        return RegistryConfig(CONFIG_FILE_NAME=config_file_name)
    except Exception as e:  # pylint: disable=bare-except
        raise WrongConfig(config_file_name) from e


CONFIG = NoFileConfig()
